#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from xoeuf import signals

__all__ = ['Enumeration']


def Enumeration(enumclass, *args, **kwargs):
    '''Create an enumeration field.

    The `enumclass` argument must be either an `enumeration class` or a
    fully qualified name of an enumeration class.

    At the DB level the enumeration will be an integer.  This requires that
    the members of provided enumeration class are instances of `int` (probably
    via a subclass).

    Enumeration classes are required to:

    - have an attribute ``__members__`` with a `collections.Mapping`:class:
      from names to values.

    - have an attribute for each name in ``__member__`` with the same value as
      in the mapping.

    Those rules imply::

       >>> all(getattr(Enumclass, name) is value and isinstance(value, int)
       ...     for name, value in Enumclass.__members__.items())

    .. note:: These requirements are compatible with Python 3.4's
       `enum.IntEnum`:class:

       Notice, however, that we don't require that members are instances of
       the enumeration class.  But it won't hurt.

    .. warning:: The returned field is a subclass of
                 `xoeuf.fields.Integer`:class: and it's a new API field.

    '''
    from xoeuf import fields
    from xoutil.objects import classproperty

    class EnumeratedField(fields.Integer, EnumerationField):
        @classproperty
        def klass(cls):
            from xoutil.objects import import_object
            return import_object(enumclass)

        @classproperty
        def members_by_value(cls):
            klass = cls.klass
            return {v: k for k, v in klass.__members__.items()}

        @classmethod
        def get_member(cls, value):
            # `value` is most likely an integer, but since we require
            # members to be subclasses of `int` they should compare equal.
            try:
                return next(v for v in cls.members_by_value if v == value)
            except StopIteration:
                raise ValueError('Invalid member of enumeration: %r' % value)

        def convert_to_cache(self, value, record, validate=True):
            value = super(EnumeratedField, self).convert_to_cache(
                value, record, validate=validate
            )
            return self.get_member(value)

        def convert_to_read(self, value, record, use_name_get=True):
            # None should be False in read.
            return self.get_member(value) if value is not None else False

    return EnumeratedField(*args, **kwargs)


class EnumerationField(object):
    '''A base clase for enumerated fields.

    '''
    pass


@signals.receiver([signals.pre_create, signals.pre_write], framework=True)
def _check_enumeration_value(sender, signal, values=None, **kwargs):
    for fieldname, value in values.items():
        field = sender._fields.get(fieldname, None)
        if isinstance(field, EnumerationField):
            # The following raises a ValueError if the value is not a member
            # of the enumeration class
            field.get_member(value)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Xœuf ORM extensions for Open Object (OpenERP) models.

All the attributes of this module must be functions that can be integrated as
`BaseModel` methods or operators.

Operators name has the format ``operator__<name>`` and are converted to
``__<name>__``.

'''


from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import)

from xoeuf.osv.writers import ORMWriter as get_writer  # noqa
orm_writer = get_writer


from xoeuf.osv.writers import ORMCreator as get_creator  # noqa
orm_creator = get_creator


def get_treeview_action(self, **options):
    '''Return the tree view action for `self`.

    There are two possible signatures:

    - ``get_treeview_action(recordset, **options)``

    If the recordset is a singleton, return the view type 'form' so that it's
    easier to see the full object.

    :rtype: An action `dict` you can return to the web client.

    `options` is used to override **almost any** key in the returned dict.  An
    interesting option is `target`.

    '''
    _, _, context = self.env.args
    ids = self.ids
    if not ids or len(ids) > 1:
        vtype = 'list'
        active_id = None
    else:
        vtype = 'form'
        active_id = ids[0]
    result = {
        'type': 'ir.actions.act_window',
        'res_model': self._name,
        'src_model': False,
        'view_id': False,
        'view_type': vtype,
        'view_mode': 'list,form',
        'views': [(False, 'list'), (False, 'form')],
        'target': 'current',
        'context': context,
    }
    if ids:
        result['domain'] = [('id', 'in', tuple(ids))]
    if active_id:
        result['active_id'] = active_id
    result.update(options)
    return result



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from . import Command


class Addons(Command):
    '''List all non-uninstallable addons found in the DB.

    '''
    @classmethod
    def get_arg_parser(cls):
        res = getattr(cls, '_arg_parser', None)
        if not res:
            from argparse import ArgumentParser
            res = ArgumentParser()
            cls._arg_parser = res
            res.add_argument('-f', '--filter',
                             dest='filters',
                             default=[],
                             action='append')
        return res

    @classmethod
    def database_factory(cls, database):
        import importlib
        module = 'xoeuf.pool.%s' % database
        return importlib.import_module(module)

    def run(self, args=None):
        parser = self.get_arg_parser()
        options = parser.parse_args(args)
        filters = options.filters
        self.list_addons(filters)

    def list_addons(self, filters):
        addons = self.get_addons(filters)
        for addon in addons:
            print(addon)

    def get_addons(self, filters):
        from xoeuf.odoo.modules import get_modules
        return [
            addon
            for addon in get_modules()
            if not filters or any(addon.startswith(f) for f in filters)
        ]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Odoo API bridge.

Eases the task of writing modules which are compatible with both Odoo and
OpenERP.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

from xoutil.deprecation import deprecated
from xoutil.decorator.meta import decorator as _xdecorator

from xoeuf.odoo import api as _odoo_api


# TODO: `copy_members` will be deprecated in xoutil 1.8, use instead the same
# mechanisms as `xoutil.future`.
from xoutil.modules import copy_members as _copy_python_module_members
this = _copy_python_module_members(_odoo_api)
del _copy_python_module_members


def contextual(func):
    '''Decorate a function to run within a proper Odoo environment.

    You should decorate every function that represents an "entry point" for
    working with the ORM.  A proper `Environment`:class: is entered upon
    calling the function.

    '''
    def inner(*args, **kwargs):
        with _odoo_api.Environment.manage():
            return func(*args, **kwargs)
    return inner


@_xdecorator
def take_one(func, index=0, warn=True, strict=False):
    '''Same as `requires_singleton()`.

    The arguments are now ignored.

    '''
    return requires_singleton(func)


_MSG = ("{funcname} is now deprecated and it will be removed. "
        "Use `{replacement}` directly and let the method raise "
        "`expected singleton` exception.")
take_one = deprecated('`api.requires_singleton()`', msg=_MSG)(take_one)
del _MSG, deprecated


def requires_singleton(f):
    '''An idiomatic alias for `api.multi()`.

    This is exactly the same as `api.multi()`, however it's expected to be
    used when the code you're decorating requires a singleton recordset.

    Notice we don't fail at the method call, but only if the actual code
    executes a command that requires such a condition to be met (for instance,
    accessing a field in ``self``.)

    '''
    return _odoo_api.multi(f)


def mimic(original):
    '''Apply the API guess of `original` to the decorated function.

    Usage::

       def f1(self, cr, uid, ids, context=None):
           # Actually any valid signature

       @api.mimic(f1)
       def f2(*args, **kwargs):
           pass

    '''
    import types
    method = _odoo_api.guess(original)
    # Odoo stores the decorator in the _api attribute.  But Odoo 10 only
    # stores the name of the API method.
    decorator = method._api
    if isinstance(decorator, types.FunctionType):
        return decorator
    else:
        return getattr(_odoo_api, decorator)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

import contextlib
from xoeuf.odoo.tests.common import TransactionCase
from xoeuf.odoo.addons.test_signals.models import (
    signals,  # this is the same implementation the models use.
    do_nothing,
    do_nothing2,
    do_nothing3,
    do_nothing_again,
    wrap_nothing,
)

from .tools import temp_attributes


class TestXoeufSignals(TransactionCase):
    at_install = False
    post_install = not at_install

    def setUp(self):
        super(TestXoeufSignals, self).setUp()
        self.Model = self.env['test_signals.signaling_model']

    @contextlib.contextmanager
    def mocks(self, method):
        def post_create(sender, signal, **kwargs):
            return method(sender, signal, **kwargs)

        def pre_create(sender, signal, **kwargs):
            return method(sender, signal, **kwargs)

        def wrapper(sender, wrapping, *args, **kwargs):
            return method(sender, wrapping, *args, **kwargs)

        mock = signals.mock_replace
        with mock(signals.post_create, method, side_effect=post_create) as post, \
             mock(signals.pre_create, method, side_effect=pre_create) as pre, \
             mock(signals.write_wrapper, method,
                  side_effect=wrapper) as wrap, \
             temp_attributes(self.env.registry, dict(ready=True)):
            yield post, pre, wrap

    def test_post_create(self):
        with self.mocks(do_nothing) as (post, pre, _), \
             self.mocks(do_nothing2) as (post2, pre2, _), \
             self.mocks(do_nothing3) as (post3, pre3, _):
            self.Model.create(dict(name='My name'))
            self.assertTrue(post.called)
            self.assertTrue(post2.called)
            self.assertTrue(post3.called)
            self.assertFalse(pre.called)
            self.assertFalse(pre2.called)
            self.assertFalse(pre3.called)

    def test_pre_create(self):
        ready = self.env.registry.ready
        with self.mocks(do_nothing_again) as (post, pre, _):
            self.Model.create(dict(name='My name'))
            self.assertFalse(post.called)
            self.assertTrue(pre.called)
        self.assertEqual(ready, self.env.registry.ready)

    def test_writer_wrap(self):
        who = self.Model.create(dict(name='My name'))
        with self.mocks(wrap_nothing) as (_, _p, wrap):
            self.assertFalse(wrap.called)
            who.write(dict(name='My new name'))
            self.assertTrue(wrap.called)
        with self.mocks(wrap_nothing) as (_, _p, wrap):
            who.name = 'My new other name'
            self.assertTrue(wrap.called)

    def test_not_called_for_partners(self):
        with self.mocks(do_nothing) as (post, pre, wrap):
            partner = self.env['res.partner'].create(dict(name='Contact Info'))
            partner.email = 'a@b.c'
            self.assertFalse(post.called and pre.called and wrap.called)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

import xoeuf.odoo.osv.fields as fields7

if not hasattr(fields7, 'timedelta'):

    def ensure_float(x):
        from datetime import timedelta
        if isinstance(x, timedelta):
            x = x.total_seconds()
        return x

    _base = fields7.float

    class timedelta(_base):
        _type = 'timedelta'

        def __init__(self, **args):
            _base.__init__(self, **args)
            _lambda = self._symbol_f
            self._symbol_f = lambda x: _lambda(ensure_float(x))
            self._symbol_set = (self._symbol_c, self._symbol_f)

        def to_field_args(self):
            raise NotImplementedError(
                "fields.timdelta is only supported in the new API."
            )

    fields7.timedelta = timedelta
    del timedelta

del fields7



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

import contextlib

from hypothesis import strategies as s, given
from xoeuf.odoo.tests.common import TransactionCase

from ..models import COLORS

colors = s.sampled_from(COLORS.__members__.values())


class TestEnum(TransactionCase):
    def setUp(self):
        super(TestEnum, self).setUp()
        self.EnumModel = self.env['test.enum.model']

    def test_can_set_valid_integers(self):
        obj = self.EnumModel.create({'color': 1})
        id = obj.id
        self.EnumModel.invalidate_cache()
        obj = self.EnumModel.browse(id)
        assert obj.color == COLORS.Red, '%r == %r' % (obj.color, COLORS.Red)

    @given(colors)
    def test_can_set_valid_values(self, color):
        obj = self.EnumModel.create({'color': color})
        id = obj.id
        self.EnumModel.invalidate_cache()
        obj = self.EnumModel.browse(id)
        assert obj.color == color, '%r == %r' % (obj.color, color)

    @given(colors)
    def test_can_set_valid_values_as_integers_and_get_values(self, color):
        obj = self.EnumModel.create({'color': int(color)})
        id = obj.id
        self.EnumModel.invalidate_cache()
        obj = self.EnumModel.browse(id)
        assert obj.color == color, '%r == %r' % (obj.color, color)
        assert isinstance(obj.color, COLORS)

    def test_cannot_set_invalid_integers(self):
        # Sinces tests are run while the registry is being populated, i.e not
        # ready, we need to trick it to allow receivers be executed.
        with force_ready(self.env.registry), self.assertRaises(ValueError):
            self.EnumModel.create({'color': 10})

    def test_cannot_write_invalid_integers(self):
        obj = self.EnumModel.create({'color': 1})
        with force_ready(self.env.registry), self.assertRaises(ValueError):
            obj.write({'color': 10})


@contextlib.contextmanager
def force_ready(registry):
    registry.ready = True
    try:
        yield
    finally:
        registry.ready = False



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Implementation for 'xoeuf.models.proxy'.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)


import re
from xoutil.objects import memoized_property

try:
    from xoeuf.odoo import models
except ImportError:
    # So that we can generate the docs easily
    pass


class ModelProxy(object):
    @memoized_property
    def _instances_(self):
        '''All the *possible* instances of this model.

        The result is an object that allows for containment tests.  The
        containment test would be equivalent to an `isinstance` check.

        Example::

           from xoeuf.models.proxy import SomeModel
           record in SomeModel._instances_

        .. warning:: This may shadow an attribute '_instances_' in the proxied
           model.

        '''
        model = self.__model

        class Instances(object):
            def __contains__(_, who):
                from xoeuf import models
                return isinstance(who, models.BaseModel) and who._name == model

        return Instances()

    def __init__(self, name):
        self.__model = _get_model(name)
        self.__env = None

    @property
    def _proxy_request(self):
        try:
            from xoeuf.odoo.http import request
            if request.env:  # An HTTP request may not be bound to a single DB.
                return request
        except RuntimeError:
            pass
        return None

    @property
    def _this(self):
        if self._proxy_request:
            return self._proxy_request.env
        import sys
        f = sys._getframe(1)
        try:
            this, tries = None, 5
            while this is None and tries and f:
                this = f.f_locals.get('self', None)
                if not isinstance(this, models.BaseModel):
                    this = cr = uid = None
                elif not hasattr(this, 'env'):
                    # We still need to support possible old-API methods
                    cr = f.f_locals.get('cr', None)
                    uid = f.f_locals.get('uid', None)
                    context = f.f_locals.get('context', None)
                    this = this.browse(cr, uid, context=context)
                f = f.f_back
                tries -= 1
            return this.env if this is not None else None
        finally:
            f = None

    def __dir__(self):
        return dir(models.BaseModel)

    def __getattr__(self, attr):
        if attr in MODULE_ATTRS:
            return globals()[attr]
        this = self._this
        if this is not None:
            return getattr(this[self.__model], attr)
        else:
            raise RuntimeError(
                'Cannot find attribute %r in proxy model. This is most '
                'likely due to an invalid call site.' % attr
            )

    def __repr__(self):
        return '<ModelProxy: %r>' % self.__model


def _get_model(name):
    if '.' in name:
        return name
    words, last = [], 0
    for i, match in enumerate(UPPERS.finditer(name)):
        pos = match.start()
        if i == 0:
            assert pos == 0, 'The first char must an upper case'
        else:
            words.append(name[last:pos].lower())
        last = pos
    words.append(name[last:].lower())
    return '.'.join(words)


UPPERS = re.compile('[A-Z]')


# Attributes which others might think all modules should have and that
# shouldn't be delegated to the underlying model.
MODULE_ATTRS = ('__file__', '__module__', )


del memoized_property



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

try:
    from xoutil.objects import temp_attributes  # TODO: migrate
except ImportError:
    from contextlib import contextmanager
    from xoutil.objects import save_attributes, smart_setter

    @contextmanager
    def temp_attributes(obj, attrs, **kwargs):
        '''A context manager that temporarily sets attributes.

        `attrs` is a dictionary containing the attributes to set.

        Keyword arguments `getter` and `setter` have the same meaning as in
        `save_attributes`:func:.  We also use the `setter` to set the values
        provided in `attrs`.

        '''
        setter = kwargs.get('setter', smart_setter)
        set_ = setter(obj)
        with save_attributes(obj, *tuple(attrs.keys()), **kwargs):
            for attr, value in attrs.items():
                set_(attr, value)
            yield

    del contextmanager



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)


from xoeuf import fields, models


class ModelA(models.TransientModel):
    _name = __name__ + '.' + 'a'

    tzone = fields.Char(default='America/Havana')
    dt = fields.Datetime()
    dt_at_tzone = fields.LocalizedDatetime


class ModelB(models.TransientModel):
    _name = __name__ + '.' + 'b'
    _inherits = {models.get_modelname(ModelA): 'a_id'}



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Profiling tools.

For development only purposes.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)


from xoutil.decorator.meta import decorator


try:
    from line_profiler import LineProfiler
    profiler = LineProfiler()

    @decorator
    def profile(func, model=True, stream=None):
        '''Decorate `func` to activate the line profiler.

        :param func: The function to decorate

        :param model: Set to False if this function is not contained in an
                      Odoo model class definition.

        :param stream:  The stream into which print the collected stats.  The
                        default is whatever LineProfiler has as default.

        '''
        from xoeuf import api
        if model:
            f = getattr(func, '_orig', func)
            wraps = api.mimic
        else:
            f = func
            wraps = lambda f: lambda x: x   # noqa

        profiler.add_function(f)

        @wraps(func)
        def inner(*args, **kwargs):
            profiler.enable_by_count()
            try:
                return f(*args, **kwargs)
            finally:
                profiler.disable_by_count()
                profiler.print_stats(stream)

        return inner

except ImportError:
    import warnings
    profiler = None

    @decorator
    def profile(func, model=None, stream=None):
        '''Decorate `func` to activate the line profiler.

        '''
        warnings.warn(
            'Trying to the profiler without the line_profiler package.  '
            'Are you trying to use the profiler in production?! '
            'Don\'t do that EVER AGAIN!!'
        )
        return func


del decorator



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)


from xoeuf.odoo.release import version_info as ODOO_VERSION_INFO
assert ODOO_VERSION_INFO >= (10, 0)
del ODOO_VERSION_INFO

import pytz

from xoeuf.odoo import fields
from xoeuf.tools import localtime_as_remotetime


class LocalizedDatetime(fields.Datetime):
    '''A field for localized datetimes.

    Localized datetimes are actually a functional field that takes two
    underlying columns a datetime and timezone name.

    :param dt: The name of the column where the datetime (in UTC) is actually
               saved.

    :param tzone:  The name of the column where the timezone where the event
                   should be presented to the user.

    Upon reading, we assume the user's timezone is properly set.

    The datetime column will be actually saved in UTC as all datetimes in
    Odoo.  But upon reading we convert to a properly shifted datetime so that
    is presented to the user in the saved time zone.

    .. note:: At the time this field is read-only, not searchable, and
              non-storable.

    '''

    _slots = {
        'dt_field': '',
        'tzone_field': '',
        'store': False,
    }

    def __init__(self, dt_field=None, tzone_field=None, **kwargs):
        # Include store=False if is not include in kwargs
        if not dt_field or not tzone_field:
            raise TypeError('LocalizedDatetime requires the surrogates fields')
        self.dt_field = dt_field
        self.tzone_field = tzone_field
        kwargs = dict(
            dict(store=False, copy=False, dt_field=dt_field,
                 tzone_field=tzone_field),
            **kwargs
        )
        super(LocalizedDatetime, self).__init__(**kwargs)

    def new(self, **kwargs):
        # Pass original args to the new one.  This ensures that the
        # tzone_field and dt_field are present.  In odoo/models.py, Odoo calls
        # this `new()` without arguments to duplicate the fields from parent
        # classes.
        kwargs = dict(self.args, **kwargs)
        return super(LocalizedDatetime, self).new(**kwargs)

    def _setup_regular_full(self, env):
        # This is to support the case where ModelB `_inherits` from a ModelA
        # with a localized datetime.  In such a case, we don't override the
        # compute method.
        super(LocalizedDatetime, self)._setup_regular_full(env)
        self.depends = tuple(
            f for f in (self.dt_field, self.tzone_field) if f
        )
        self.compute = self._compute
        if not self.readonly:
            self.inverse = self._inverse
        self.search = self._search

    def _compute(self, records):
        tzone_field = self.tzone_field
        dt_field = self.dt_field
        tz = records._context.get('tz', None)
        if not tz:
            user = records.env.user
            tz = pytz.timezone(user.tz) if user.tz else pytz.UTC
        else:
            tz = pytz.timezone(tz)
        for item in records:
            tzone = getattr(item, tzone_field)
            if not tzone:
                tzone = pytz.UTC
            else:
                tzone = pytz.timezone(tzone)
            dt = getattr(item, dt_field)
            # Compute the datetime in users timezone,
            # then force to it the desired TZ and back to UTC.
            if dt and tz != tzone:
                dt = localtime_as_remotetime(dt, tzone, tz)
            setattr(item, self.name, dt)

    def _inverse(self, records):
        tzone_field = self.tzone_field
        dt_field = self.dt_field
        tz = records._context.get('tz', None)
        if not tz:
            user = records.env.user
            tz = pytz.timezone(user.tz) if user.tz else pytz.UTC
        else:
            tz = pytz.timezone(tz)
        for item in records:
            tzone = getattr(item, tzone_field)
            if not tzone:
                tzone = pytz.UTC
            else:
                tzone = pytz.timezone(tzone)
            dt = getattr(item, self.name)
            # Compute the datetime in the desired timezone, then
            # extract all datetime components but the TZ and localize
            # it to the users TZ and convert it back to UTC...
            # This makes the UI to reverse the process and show the
            # datetime in the desired timezone.
            if dt and tz != tzone:
                dt = localtime_as_remotetime(dt, tz, tzone)
            setattr(item, dt_field, dt)

    def _search(self, records, operator, value):
        # TODO: localize value.
        return [(self.dt_field, operator, value)]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''This is not a future extension module.

Add a function to assure that future module is being used correctly.

'''

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)


def dissuade(name=None):
    '''Check future module for being used correctly.

    If not, -meaning in the past- issue a deprecation warning.

    :param name: name of module to check (could be ``None`` to inspect it in
           the stack-frame).

    '''
    if name is None:
        import inspect
        frame = inspect.currentframe().f_back
        try:
            name = frame.f_globals.get('__name__')
        finally:
            # As recommended in Python's documentation to avoid memory leaks
            del frame
    prefix = '.'.join(__name__.split('.')[:-1])
    if name and prefix not in name:
        from warnings import warn
        new_name = '{}.{}'.format(prefix, name.split('.')[-1])
        warn(('"{}" is now deprecated and it will be removed; use "{}" '
              'instead.').format(name, new_name),
             # TODO: Why ``category=DeprecationWarning,`` doesn't work?
             stacklevel=3)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)

import unittest


class TestModules(unittest.TestCase):
    def test_get_object_module(self):
        from .. import Foo
        from xoeuf.modules import get_object_module
        self.assertEqual(get_object_module(Foo), Foo.module)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

'''Xœuf basic ORM extensions for Open Object (OpenERP) models.'''


from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import)

from xoeuf.models import get_modelname as _get_modelname
from xoutil.deprecation import deprecated

get_modelname = deprecated('xoeuf.models.get_modelname')(_get_modelname)
del _get_modelname, deprecated


def guess_id(which, attr='id'):
    '''Guess the id of an object.

    If `which` is an integer, it is returned unchanged.  If it is a dict
    or a browse_record the attribute/key given by `attr` is look up and
    return.  If not found an AttibuteError is raised.  Any other type is a
    TypeError.

    '''
    from xoeuf.odoo.osv.orm import browse_record
    from xoutil.future.collections import Mapping
    from xoutil.eight import integer_types
    if isinstance(which, integer_types):
        return which
    elif isinstance(which, (Mapping, browse_record)):
        from xoutil.objects import smart_getter
        get = smart_getter(which, strict=True)
        return get(attr)
    else:
        raise TypeError('Object %r of invalid type' % which)


def store_identity(self, cr, uid, ids, context=None):
    '''To be used in ``store`` parameter for functional fields when monitor
    is needed for fields in a local model.

    For example::

        'active':
            fields.function(_get_active, type='boolean', string='Active?',
                store={_name: (store_identity, ['contract_ids'], 10), }),

    '''
    return ids


# Making this a subclass of int allows to be reliably tested for equality in
# both Python 2.7 and Python 3.4
class _Command(int):
    def __new__(cls, value, *args):
        return super(_Command, cls).__new__(cls, value)

    def __init__(self, value, _lambda):
        self.value = value
        self._lambda = _lambda

    def __call__(self, *args, **kwargs):
        if callable(self._lambda):
            return self._lambda(*args, **kwargs)
        else:
            return self._lambda


# The index where the command magic number is in tuples
COMMAND_INDEX = 0

# For applicable commands this is the index where id/ids are located.
ID_INDEX = 1

# For applicable commands this is the index where values are located
VALUE_INDEX = VALUES_INDEX = -1


#: Returns a single "command" to create a new related record
CREATE_RELATED = _Command(0, lambda **values: (0, 0, values))
ONE2MANY_CREATE = MANY2MANY_CREATE = CREATE_RELATED

#: Returns a single "command" to update a linked record
UPDATE_RELATED = _Command(1, lambda id, **values: (1, id, values))
ONE2MANY_UPDATE = MANY2MANY_UPDATE = UPDATE_RELATED

#: Returns a single "command" to remove (and unlink) the related record
REMOVE_RELATED = _Command(2, lambda id: (2, id))
ONE2MANY_REMOVE = MANY2MANY_REMOVE = REMOVE_RELATED

#: Returns a single command to forget about a relation
FORGET_RELATED = _Command(3, lambda id: (3, id))

#: Returns a single command to link a record
LINK_RELATED = _Command(4, lambda id: (4, id))

#: Returns a single command to unlink all
UNLINKALL_RELATED = _Command(5, lambda: (5, ))

#: Returns a single command to replace all related with existing ids
REPLACEWITH_RELATED = _Command(6, lambda *ids: (6, 0, list(ids)))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

from __future__ import (division as _py3_division,
                        print_function as _py3_print,
                        absolute_import as _py3_abs_import)


from xoeuf.odoo.tests.common import TransactionCase


class TestProperty(TransactionCase):
    def setUp(self):
        super(TestProperty, self).setUp()
        Value = self.Value = self.env['test.property.value']
        self.obj = Value.create({'value': '0123456789'})
        for val in 'abcdefg':
            Value.create({'value': val})

    def test_setup(self):
        self.assertTrue(getattr(type(self.Value), 'inverted_setup', False))

    def test_getter(self):
        obj = self.obj
        self.assertEqual(obj.inverted, '9876543210')

    def test_setter(self):
        obj = self.obj
        obj.inverted = 'abcd'
        self.assertEqual(obj.value, 'dcba')

    def test_deleter(self):
        obj = self.obj
        del obj.inverted
        self.assertIn(obj.value, (None, False))  # Odoo may xform None to False

    def test_nocreate(self):
        with self.assertRaises(AssertionError):
            # I would like that calling create would raise an exception, but
            # for now it just does not do anything.
            obj = self.Value.create({'inverted': 'abcd'})
            self.assertEqual(obj.value, 'dcba')

    def test_nowrite(self):
        with self.assertRaises(AssertionError):
            # I would like that calling search would raise an exception, but
            # for now it just does not do anything.
            self.obj.write({'inverted': 'abcd'})
            self.assertEqual(self.obj.value, 'dcba')

    # Don't perform the test of `search()` because it does not raise an
    # exception, but log an error that may fool our test scripts.


class TestInheritedValue(TransactionCase):
    def setUp(self):
        super(TestInheritedValue, self).setUp()
        Value = self.Value = self.env['test.property.value']
        Identifier = self.Identifier = self.env['test.property.inherits']
        for i in range(10):
            val = Value.create({'thing': repr({i: i})})
            iden = Identifier.create({'name': 'v%d' % i, 'value_id': val.id})
            self.assertEqual(val.thing, iden.thing)

    def test_result_write(self):
        i = self.Identifier.search([], limit=1)
        v = {'a': [1, (1, 2), u'a']}
        i.write({'result': v})
        self.assertEqual(i.thing, repr(v))
